#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import itertools
import math
import pytest

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentMap, ComponentSet
import pyomo.environ as pyo
import pyomo.dae as dae
from pyomo.core.expr.visitor import identify_variables
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.dae.flatten import flatten_dae_components

from pyomo.contrib.incidence_analysis import (
    IncidenceGraphInterface,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.util.subsystems import ParamSweeper

import idaes.core as idaes
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    large_residuals_set,
)
from idaes.core.util.constants import Constants
from idaes.models_extra.gas_distribution.properties.natural_gas import (
    NaturalGasParameterBlock,
)
from idaes.models_extra.gas_distribution.unit_models.pipeline import GasPipeline

from idaes.apps.nmpc import (
    get_tracking_cost_from_constant_setpoint as get_tracking_cost_expression,
)
from idaes.apps.nmpc.dynamic_data import (
    load_inputs_into_model,
    interval_data_from_time_series,
)
from idaes.core.solvers import get_solver

"""
Test for the simple pipeline model
"""


@pytest.mark.component
class TestSolvePipelineSquare(unittest.TestCase):
    """ """

    # These dictionaries map parameters in a steady state pipeline
    # solve to the predicted values for outlet flow rate and pressure.
    # These were obtained by solving a Pyomo DAE implementation of the
    # same model.
    target_flow_kghr = {
        # flow key is also in units of kg/hr. Should this be in
        # 1e4 SCM/hr, which is what the paper uses?
        (180000.0, 40.0, 1): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 40.0, 2): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 40.0, 5): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 57.0, 1): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 57.0, 2): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 57.0, 5): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 70.0, 1): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 70.0, 2): 180000.0 * pyo.units.kg / pyo.units.hr,
        (180000.0, 70.0, 5): 180000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 40.0, 1): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 40.0, 2): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 40.0, 5): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 57.0, 1): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 57.0, 2): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 57.0, 5): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 70.0, 1): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 70.0, 2): 270000.0 * pyo.units.kg / pyo.units.hr,
        (270000.0, 70.0, 5): 270000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 40.0, 1): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 40.0, 2): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 40.0, 5): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 57.0, 1): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 57.0, 2): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 57.0, 5): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 70.0, 1): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 70.0, 2): 450000.0 * pyo.units.kg / pyo.units.hr,
        (450000.0, 70.0, 5): 450000.0 * pyo.units.kg / pyo.units.hr,
    }
    target_pressure = {
        (180000.0, 40.0, 1): 37.00032465199001 * pyo.units.bar,
        (180000.0, 40.0, 2): 36.94189598799329 * pyo.units.bar,
        (180000.0, 40.0, 5): 36.90451650814217 * pyo.units.bar,
        (180000.0, 57.0, 1): 54.89496466806317 * pyo.units.bar,
        (180000.0, 57.0, 2): 54.87516406405006 * pyo.units.bar,
        (180000.0, 57.0, 5): 54.86291547799542 * pyo.units.bar,
        (180000.0, 70.0, 1): 68.28589980113715 * pyo.units.bar,
        (180000.0, 70.0, 2): 68.27527627682545 * pyo.units.bar,
        (180000.0, 70.0, 5): 68.26877313026228 * pyo.units.bar,
        (270000.0, 40.0, 1): 33.25073046697753 * pyo.units.bar,
        (270000.0, 40.0, 2): 32.939794050787654 * pyo.units.bar,
        (270000.0, 40.0, 5): 32.72133469851969 * pyo.units.bar,
        (270000.0, 57.0, 1): 52.26367050314213 * pyo.units.bar,
        (270000.0, 57.0, 2): 52.16101599318696 * pyo.units.bar,
        (270000.0, 57.0, 5): 52.09485138003043 * pyo.units.bar,
        (270000.0, 70.0, 1): 66.1432745525586 * pyo.units.bar,
        (270000.0, 70.0, 2): 66.0886470609479 * pyo.units.bar,
        (270000.0, 70.0, 5): 66.05431808504241 * pyo.units.bar,
        (450000.0, 40.0, 1): 21.252029074937592 * pyo.units.bar,
        (450000.0, 40.0, 2): 18.382847277082647 * pyo.units.bar,
        (450000.0, 40.0, 5): 15.152518801949473 * pyo.units.bar,
        (450000.0, 57.0, 1): 43.8435291753948 * pyo.units.bar,
        (450000.0, 57.0, 2): 42.98530492454757 * pyo.units.bar,
        (450000.0, 57.0, 5): 42.33711761875827 * pyo.units.bar,
        (450000.0, 70.0, 1): 59.286873757107195 * pyo.units.bar,
        (450000.0, 70.0, 2): 58.84301170614686 * pyo.units.bar,
        (450000.0, 70.0, 5): 58.53638005215126 * pyo.units.bar,
    }

    def test_pipeline_steady_forward_nominal(self):
        """
        Test forward simulation of the pipeline (flow and pressure
        specified at the inlet) with nominal inlet conditions at
        steady state. Here just test that we can solve with no errors.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 1,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline
        p = "Vap"
        j = next(iter(m.fs.properties.component_list))
        # Fix pipeline degrees of freedom
        # Design variables:
        pipeline.diameter.fix(0.92 * pyo.units.m)
        # Pipelines in paper are 300 km. This seems long. Is this really
        # the length we're interested in?
        pipeline.control_volume.length.fix(300.0 * pyo.units.km)
        # Inlet variables:
        x0 = pipeline.control_volume.length_domain.first()
        state = pipeline.control_volume.properties
        state[:, x0].mole_frac_comp[j].fix(1.0)
        state[:, x0].temperature.fix(293.15 * pyo.units.K)
        state[:, x0].pressure.fix(57.0 * pyo.units.bar)
        pipeline.control_volume.flow_mass[:, x0].fix(
            3.0e5
            * pyo.units.kg
            / pyo.units.hr
            # close to 10 * (1e6 SCM) / day, the nominal value in the model
        )

        t0 = m.fs.time.first()

        ipopt = get_solver("ipopt")

        res = ipopt.solve(m, tee=True)
        pyo.assert_optimal_termination(res)

        violated_cons = list(large_residuals_set(m))
        # Just test that we can solve the pipeline model.
        # We will test values in the next test.
        self.assertEqual(len(violated_cons), 0)

    def test_pipeline_steady_forward_sweep(self):
        """
        Tests "forward simulation" (supply flow and pressure specified)
        of the steady state pipeline over a range of input values and
        numbers of spatial finite elements.
        """
        nxfe_list = [1, 2, 5]
        p_list = [40.0 * pyo.units.bar, 57.0 * pyo.units.bar, 70.0 * pyo.units.bar]
        h_day = 24.0 * pyo.units.hr / pyo.units.day
        kg_scm = 0.72 * pyo.units.kg / pyo.units.m**3
        f_list = [
            # Units: kg/hr
            (6.0 * 1e6) * pyo.units.m**3 / pyo.units.day / h_day * kg_scm,
            (9.0 * 1e6) * pyo.units.m**3 / pyo.units.day / h_day * kg_scm,
            (15.0 * 1e6) * pyo.units.m**3 / pyo.units.day / h_day * kg_scm,
        ]
        # Iterate over nxfe first, because for each different nxfe, we
        # need to create a new model.

        for nxfe in nxfe_list:
            m = pyo.ConcreteModel()
            default = {
                "dynamic": False,
            }
            m.fs = idaes.FlowsheetBlock(**default)
            m.fs.properties = NaturalGasParameterBlock()
            pipeline_config = {
                "property_package": m.fs.properties,
                "finite_elements": nxfe,
            }
            m.fs.pipeline = GasPipeline(**pipeline_config)
            pipeline = m.fs.pipeline
            p = "Vap"
            j = next(iter(m.fs.properties.component_list))
            # Fix pipeline degrees of freedom
            # Design variables:
            pipeline.diameter.fix(0.92 * pyo.units.m)
            # Pipelines in paper are 300 km. This seems long. Is this really
            # the length we're interested in?
            pipeline.control_volume.length.fix(300.0 * pyo.units.km)
            # Inlet variables:
            x0 = pipeline.control_volume.length_domain.first()
            xf = pipeline.control_volume.length_domain.last()
            state = pipeline.control_volume.properties
            state[:, x0].mole_frac_comp[j].fix(1.0)
            state[:, x0].temperature.fix(293.15 * pyo.units.K)
            state[:, x0].pressure.fix()
            pipeline.control_volume.flow_mass[:, x0].fix()
            t0 = m.fs.time.first()

            input_values = ComponentMap()
            target_values = ComponentMap()
            t0 = m.fs.time.first()
            inlet_flow = pipeline.control_volume.flow_mass[t0, x0]
            inlet_pressure = state[t0, x0].pressure
            outlet_flow = pipeline.control_volume.flow_mass[t0, xf]
            outlet_pressure = state[t0, xf].pressure
            input_values[inlet_flow] = []
            input_values[inlet_pressure] = []
            target_values[outlet_flow] = []
            target_values[outlet_pressure] = []
            for f, p in itertools.product(f_list, p_list):
                # Maybe I should explicitly convert to desired units here,
                # for clarity.
                input_values[inlet_flow].append(f)
                # p_val = pyo.units.convert(p, pres_var.get_units())
                input_values[inlet_pressure].append(p)
                # Evaluating floating points as keys here may not be safe...
                target_values[outlet_flow].append(
                    self.target_flow_kghr[pyo.value(f), pyo.value(p), nxfe],
                )
                target_values[outlet_pressure].append(
                    self.target_pressure[pyo.value(f), pyo.value(p), nxfe]
                )

            n_scen = len(f_list) * len(p_list)
            param_sweeper = ParamSweeper(
                n_scen,
                input_values,
                to_fix=list(input_values.keys()),
                output_values=target_values,
            )
            ipopt = get_solver("ipopt")
            with param_sweeper:
                # Note that doing this in a context manager means that
                # on error, values are reset. This is inconvenient
                # for debugging.
                # NOTE: Entering post-mortem debugging appears to
                # exit the context manager before returning control...
                self.assertEqual(degrees_of_freedom(m), 0)
                for i, (inputs, outputs) in enumerate(param_sweeper):
                    res = ipopt.solve(m, tee=True)
                    # Check for optimal termination and zero infeasibility
                    pyo.assert_optimal_termination(res)
                    self.assertEqual(len(large_residuals_set(m)), 0)

                    # Sanity check that inputs have been properly set.
                    for var, val in inputs.items():
                        val = pyo.value(pyo.units.convert(val, var.get_units()))
                        self.assertEqual(val, var.value)

                    for var, val in outputs.items():
                        val = pyo.value(pyo.units.convert(val, var.get_units()))
                        # Assume var is (or has been scaled to be) in the
                        # same units as the target value.
                        val = pyo.value(val)
                        self.assertTrue(math.isclose(val, var.value, rel_tol=1e-2))


@pytest.mark.component
class TestSolveDynamicPipeline(unittest.TestCase):
    def make_steady_model(
        self,
        nfe=2,
        scheme="FORWARD",
    ):
        m = pyo.ConcreteModel()
        default = {"dynamic": False}
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": nfe,
            "transformation_scheme": scheme,
            "has_holdup": True,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        cv = m.fs.pipeline.control_volume
        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)
        return m

    def fix_model_inlets(
        self,
        model,
        inlet_flow_mass=3.0e5 * pyo.units.kg / pyo.units.hr,
        inlet_pressure=57.0 * pyo.units.bar,
        inlet_temperature=293.15 * pyo.units.K,
    ):
        cv = model.fs.pipeline.control_volume
        j = next(iter(model.fs.properties.component_list))
        x0 = cv.length_domain.first()
        cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix(inlet_temperature)
        cv.pressure[:, x0].fix(inlet_pressure)
        cv.flow_mass[:, x0].fix(inlet_flow_mass)

    def get_scalar_data_from_model(
        self,
        model,
        time,
        scalar_vars=None,
        dae_vars=None,
    ):
        if scalar_vars is None or dae_vars is None:
            scalar_vars, dae_vars = flatten_dae_components(model, time, pyo.Var)
        return {str(pyo.ComponentUID(var)): var.value for var in scalar_vars}

    def get_data_from_model_at_time(
        self,
        model,
        time,
        scalar_vars=None,
        dae_vars=None,
        t0=0,
    ):
        if scalar_vars is None or dae_vars is None:
            scalar_vars, dae_vars = flatten_dae_components(model, time, pyo.Var)
        return {str(pyo.ComponentUID(var.referent)): var[t0].value for var in dae_vars}

    def test_sim_inlet_pressure_outlet_flow_nominal(self):
        """
        Inlet pressure and outlet flow rate will be fixed.
        """
        nxfe = 4
        ipopt = get_solver("ipopt")

        m_steady = self.make_steady_model(nfe=nxfe)
        self.fix_model_inlets(m_steady)
        ipopt.solve(m_steady, tee=True)
        time_steady = m_steady.fs.time
        scalar_data = self.get_scalar_data_from_model(m_steady, time_steady)
        initial_data = self.get_data_from_model_at_time(m_steady, time_steady)

        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": nxfe,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        cv = m.fs.pipeline.control_volume
        assert_units_consistent(m)

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 20
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        time = m.fs.time
        t0 = m.fs.time.first()
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        j = next(iter(m.fs.properties.component_list))

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Fix dynamic inputs. Now these are inlet pressure and outlet
        # flow rate, as well as inlet mole fraction and temperature.
        cv.pressure[:, x0].fix()
        cv.flow_mass[:, xf].fix()
        cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix()

        # Fix initial conditions. Here, pressure and volume for all
        # non-specified points.
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        self.assertEqual(degrees_of_freedom(m), 0)

        # Load initial steady state into model at all time points.
        for name, val in initial_data.items():
            var = m.find_component(name)
            for t in time:
                var[t].set_value(val)
        # Load scalar data from initial steady state
        # (initialize area, basically)
        for name, val in scalar_data.items():
            var = m.find_component(name)
            var.set_value(val)

        cv.material_accumulation[...].set_value(0.0)
        cv.flow_mass_dt[...].set_value(0.0)

        for con in large_residuals_set(m):
            resid = pyo.value(con.body - con.upper)
            print(resid, con.name)
        ipopt.solve(m, tee=True)

        # Load input sequence into model
        sample_points = [4.0, 20.0]
        input_name = "fs.pipeline.control_volume.flow_mass[*,1.0]"
        nominal_density = 0.72
        val = 12.0 * 1e6 / 24 * nominal_density  # 12 (1e6 SCM)/day
        input_series_data = (
            sample_points,
            {input_name: [val, val]},
        )
        input_interval_data = interval_data_from_time_series(input_series_data)
        load_inputs_into_model(m, time, input_interval_data)
        # Solve with loaded inputs
        res = ipopt.solve(m, tee=True)
        self.assertIs(
            res.solver.termination_condition,
            pyo.TerminationCondition.optimal,
        )

        # These predicted values come from a simulation of a single pipeline
        # model from the Pyomo DAE example. flow_mass has been converted
        # to kg/hr from (1e4 SCM/hr) by a factor of 0.72*1e4, where
        # 0.72 kg/m**3 is the gas density at standard conditions.
        pred_values = (
            list(time),
            {
                "fs.pipeline.control_volume.flow_mass[*,%s]"
                % x0: [
                    3.000e5,
                    2.999e5,
                    2.999e5,
                    2.999e5,
                    3.000e5,
                    3.174e5,
                    3.301e5,
                    3.389e5,
                    3.449e5,
                    3.492e5,
                    3.523e5,
                    3.544e5,
                    3.560e5,
                    3.571e5,
                    3.579e5,
                    3.585e5,
                    3.589e5,
                    3.592e5,
                    3.594e5,
                    3.595e5,
                    3.597e5,
                ],
                "fs.pipeline.control_volume.pressure[*,%s]"
                % xf: [
                    50.90,
                    50.90,
                    50.90,
                    50.90,
                    50.90,
                    49.83,
                    49.31,
                    48.95,
                    48.69,
                    48.51,
                    48.38,
                    48.29,
                    48.22,
                    48.17,
                    48.14,
                    48.11,
                    48.10,
                    48.08,
                    48.07,
                    48.07,
                    48.06,
                ],
            },
        )
        output_names = [
            "fs.pipeline.control_volume.flow_mass[*,%s]" % x0,
            "fs.pipeline.control_volume.pressure[*,%s]" % xf,
        ]
        actual_values = (
            list(time),
            {
                name: [var.value for var in m.find_component(name).values()]
                for name in output_names
            },
        )
        # Note: We fail with a reltol of 0.01, due to flow rate discrepancies
        # in positions 6, 7, 8, and 9. A reltol of 0.02 seems reasonable to me.
        self.assertStructuredAlmostEqual(pred_values, actual_values, reltol=0.02)

    def test_opt(self):
        """
        A dynamic optimization problem with a 20 hr horizon. Inputs are inlet
        pressure, which is restricted to be piecewise constant,
        and outlet flow rate, which is unrestricted.
        The setpoint is the steady state of the system with inlets of 57 bar
        and 5e5 kg/hr.
        """
        nxfe = 4
        ipopt = get_solver("ipopt")

        # Steady state data
        m_steady = self.make_steady_model(nfe=nxfe)
        self.fix_model_inlets(m_steady)
        ipopt.solve(m_steady, tee=True)
        time_steady = m_steady.fs.time
        scalar_data = self.get_scalar_data_from_model(m_steady, time_steady)
        initial_data = self.get_data_from_model_at_time(m_steady, time_steady)
        #

        # Setpoint data
        m_setpoint = self.make_steady_model(nfe=nxfe)
        self.fix_model_inlets(
            m_setpoint,
            inlet_flow_mass=5.0e5 * pyo.units.kg / pyo.units.hr,
        )
        ipopt.solve(m_setpoint, tee=True)
        time_setpoint = m_setpoint.fs.time
        setpoint_data = self.get_data_from_model_at_time(m_setpoint, time_setpoint)
        #

        t0 = 0.0
        horizon = 20.0
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [t0, horizon],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": nxfe,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        cv = m.fs.pipeline.control_volume
        assert_units_consistent(m)

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 20
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        time = m.fs.time
        t0 = m.fs.time.first()
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        j = next(iter(m.fs.properties.component_list))

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Fix "dynamic parameters" in the optimization problem
        cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix()

        # Fix initial conditions. Here, pressure and volume except
        # at points where we treat them as "inputs"
        for x in cv.length_domain:
            # For dynamic optimization, I believe we want these variables
            # fixed everywhere
            cv.pressure[t0, x].fix()
            cv.flow_mass[t0, x].fix()

        # Now we set up the dynamic optimization problem:
        tracking_variables = [
            # NOTE: These space keys (0.0 and 1.0) need to be consistent with
            # how they are declared in a continuous set, since we are using
            # the CUID string representation as a key in setpoint_data.
            pyo.Reference(cv.pressure[:, 0.0]),
            pyo.Reference(cv.pressure[:, 1.0]),
            pyo.Reference(cv.flow_mass[:, 0.0]),
            pyo.Reference(cv.flow_mass[:, 1.0]),
        ]
        weight_data = {
            "fs.pipeline.control_volume.flow_mass[*,0.0]": 1e-10,
            "fs.pipeline.control_volume.flow_mass[*,1.0]": 1e-10,
            "fs.pipeline.control_volume.pressure[*,0.0]": 1e-2,
            "fs.pipeline.control_volume.pressure[*,1.0]": 1e-2,
        }
        m.tracking_cost = get_tracking_cost_expression(
            tracking_variables, time, setpoint_data, weight_data=weight_data
        )
        m.tracking_objective = pyo.Objective(expr=sum(m.tracking_cost.values()))

        sample_period = 2  # hours
        n_samples = (int(horizon) - int(t0)) // sample_period
        sample_points = [t0 + float(sample_period * i) for i in range(n_samples + 1)]
        sample_point_set = set(sample_points)
        piecewise_constant_vars = [pyo.Reference(cv.pressure[:, 0.0])]
        m.piecewise_constant_vars_set = pyo.Set(
            initialize=list(range(len(piecewise_constant_vars)))
        )

        def piecewise_constant_rule(m, i, t):
            var = piecewise_constant_vars[i]
            if t in sample_point_set:
                return pyo.Constraint.Skip
            else:
                t_next = time.next(t)
                return var[t] == var[t_next]

        m.piecewise_constant_constraint = pyo.Constraint(
            m.piecewise_constant_vars_set, time, rule=piecewise_constant_rule
        )
        self.assertEqual(
            degrees_of_freedom(m),
            # Pressure has n_samples dof, flow has (len(time)-1) dof
            n_samples + (len(time) - 1),
        )

        # Initialize the dynamic optimization problem:

        # Load initial steady state into model at all time points.
        for name, val in initial_data.items():
            var = m.find_component(name)
            for t in time:
                var[t].set_value(val)
        # Load scalar data from initial steady state
        # (initialize area, basically)
        for name, val in scalar_data.items():
            var = m.find_component(name)
            var.set_value(val)

        cv.material_accumulation[...].set_value(0.0)
        cv.flow_mass_dt[...].set_value(0.0)

        self.assertEqual(len(large_residuals_set(m)), 0)

        res = ipopt.solve(m, tee=True)
        self.assertIs(
            res.solver.termination_condition,
            pyo.TerminationCondition.optimal,
        )

        # These predicted values were calculated from a single-pipeline
        # model created from the Pyomo DAE example. That model uses the
        # same tracking objective used here, but with flow rate weights
        # scaled to match the different units in that model (1e4 SCM)/hr.
        # The values obtained from that model are then converted back
        # to kg/hr (via the standard density 0.72 kg/m^3).
        pred_tracking_variable_data = [
            [
                57.00,
                58.86,
                58.86,
                58.71,
                58.71,
                58.35,
                58.35,
                58.04,
                58.04,
                57.80,
                57.80,
                57.63,
                57.63,
                57.51,
                57.51,
                57.42,
                57.42,
                57.38,
                57.38,
                57.35,
                57.35,
            ],
            [
                50.90,
                45.89,
                43.59,
                42.27,
                41.50,
                40.97,
                40.62,
                40.31,
                40.09,
                39.88,
                39.72,
                39.56,
                39.45,
                39.34,
                39.26,
                39.18,
                39.13,
                39.08,
                39.05,
                39.02,
                39.01,
            ],
            [
                3.000e5,
                4.693e5,
                4.629e5,
                4.698e5,
                4.850e5,
                4.761e5,
                4.909e5,
                4.819e5,
                4.936e5,
                4.866e5,
                4.953e5,
                4.903e5,
                4.966e5,
                4.932e5,
                4.977e5,
                4.955e5,
                4.985e5,
                4.973e5,
                4.992e5,
                4.986e5,
                4.996e5,
            ],
            [
                3.000e5,
                5.550e5,
                5.456e5,
                5.362e5,
                5.294e5,
                5.261e5,
                5.212e5,
                5.194e5,
                5.158e5,
                5.144e5,
                5.118e5,
                5.107e5,
                5.087e5,
                5.078e5,
                5.063e5,
                5.056e5,
                5.044e5,
                5.038e5,
                5.029e5,
                5.022e5,
                5.013e5,
            ],
        ]
        pred_values = (
            list(time),
            {
                str(pyo.ComponentUID(var.referent)): values
                for var, values in zip(tracking_variables, pred_tracking_variable_data)
            },
        )

        actual_values = (
            list(time),
            {
                str(pyo.ComponentUID(var.referent)): [var[t].value for t in time]
                for var in tracking_variables
            },
        )
        self.assertStructuredAlmostEqual(pred_values, actual_values, reltol=0.02)


@pytest.mark.unit
class TestConstructPipeline(unittest.TestCase):
    """
    Test for construction of pipeline
    """

    def test_geometry(self):
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        pipeline = m.fs.pipeline
        cv = pipeline.control_volume
        self.assertTrue(isinstance(cv.area, pyo.Var))
        self.assertTrue(isinstance(m.fs.pipeline.diameter, pyo.Var))

        diameter = m.fs.pipeline.diameter
        diameter_eqn = m.fs.pipeline.diameter_eqn
        cv.area.fix(5.0 * pyo.units.m**2)
        calculate_variable_from_constraint(diameter, diameter_eqn)
        self.assertAlmostEqual(
            diameter.value,
            pyo.value(
                pyo.units.convert(
                    pyo.sqrt(cv.area / Constants.pi) * 2, diameter.get_units()
                )
            ),
        )

    def test_momentum_balance_steady(self):
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        pipeline = m.fs.pipeline
        cv = pipeline.control_volume

        self.assertTrue(isinstance(cv.momentum_balance, pyo.Constraint))
        self.assertTrue(isinstance(cv.pressure_dx, dae.DerivativeVar))
        self.assertTrue(isinstance(cv.pressure_dx_disc_eq, pyo.Constraint))
        assert_units_consistent(cv.momentum_balance)

        t = m.fs.time.first()
        x = m.fs.pipeline.control_volume.length_domain.first()

        momentum_bal = cv.momentum_balance[t, x]
        var_set = ComponentSet(identify_variables(momentum_bal.expr))
        self.assertIn(cv.pressure_dx[t, x], var_set)

    def test_steady_forward(self):
        """
        Tests a steady state instance of the pipeline for forward simulation
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": False,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)
        pipeline = m.fs.pipeline

        cv = m.fs.pipeline.control_volume
        self.assertTrue(isinstance(cv.material_balances, pyo.Constraint))
        self.assertTrue(isinstance(cv.material_flow_dx, dae.DerivativeVar))
        self.assertTrue(isinstance(cv.length_domain, dae.ContinuousSet))
        self.assertTrue(isinstance(cv.material_flow_dx_disc_eq, pyo.Constraint))
        self.assertTrue(isinstance(cv.momentum_balance, pyo.Constraint))
        self.assertTrue(isinstance(cv.pressure_dx, dae.DerivativeVar))
        p = "Vap"
        j = next(iter(m.fs.properties.component_list))
        t = m.fs.time.first()
        x = cv.length_domain.first()
        flux = cv.material_flow_dx[t, x, p, j]
        mat_bal = cv.material_balances[t, x, p, j]
        var_set = ComponentSet(identify_variables(mat_bal.expr))
        self.assertIn(flux, var_set)
        # Really would like to assert that mat bal only has one nonzero term
        # and is linear wrt flow_dx

        # Fix pipeline degrees of freedom
        # Design variables:
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        m.fs.pipeline.control_volume.length.fix(300.0 * pyo.units.m)
        # Inlet variables:
        x0 = m.fs.pipeline.control_volume.length_domain.first()
        xf = m.fs.pipeline.control_volume.length_domain.last()
        state = m.fs.pipeline.control_volume.properties
        state[:, x0].mole_frac_comp[j].fix(1.0)
        state[:, x0].temperature.fix(300.0 * pyo.units.K)
        state[:, x0].pressure.fix(57.0 * pyo.units.bar)
        cv.flow_mass[:, x0].fix(
            3.0e5
            * pyo.units.kg
            / pyo.units.hr
            # close to 10 * (1e6 SCM) / day, the nominal value in the model
        )

        # This test asserts:
        # (a) consistent units
        # (b) zero degrees of freedom
        # (c) a perfect matching between constraints and variables
        assert_units_consistent(m)
        self.assertEqual(degrees_of_freedom(m), 0)
        igraph = IncidenceGraphInterface(m)
        matching = igraph.maximum_matching()
        N, M = igraph.incidence_matrix.shape
        self.assertEqual(N, M)
        self.assertEqual(len(matching), N)

    def test_dynamic_forward(self):
        """
        Tests construction of a dynamic instance of the pipeline for
        forward simulation.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 4,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        cv = m.fs.pipeline.control_volume
        self.assertTrue(isinstance(cv.material_balances, pyo.Constraint))
        self.assertTrue(isinstance(cv.material_flow_dx, dae.DerivativeVar))
        self.assertTrue(isinstance(cv.material_accumulation, dae.DerivativeVar))
        self.assertTrue(isinstance(cv.length_domain, dae.ContinuousSet))
        self.assertTrue(isinstance(cv.material_flow_dx_disc_eq, pyo.Constraint))
        p = "Vap"
        j = next(iter(m.fs.properties.component_list))
        t = m.fs.time.first()
        x = cv.length_domain.first()
        flux = cv.material_flow_dx[t, x, p, j]
        accum = cv.material_accumulation[t, x, p, j]
        mat_bal = cv.material_balances[t, x, p, j]
        var_set = ComponentSet(identify_variables(mat_bal.expr))
        self.assertIn(flux, var_set)
        self.assertIn(accum, var_set)
        assert_units_consistent(mat_bal)
        assert_units_consistent(m)

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 5
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        t0 = m.fs.time.first()
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Fix dynamic inputs. Here, inlet flow and pressure for all time.
        cv.pressure[:, x0].fix()
        cv.flow_mass[:, x0].fix()
        cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix()

        # Fix initial conditions. Here, pressure and volume for all
        # non-inlet points.
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
                cv.flow_mass[t0, x].fix()

        igraph = IncidenceGraphInterface(m)
        N, M = igraph.incidence_matrix.shape
        matching = igraph.maximum_matching()
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)  # Sanity check
        self.assertEqual(len(matching), N)

    def test_dynamic_inlet_pressure_outlet_flow(self):
        """
        Dynamic simulation with both inlets fixed appears to be unstable,
        so we'll proceed with inlet pressure and outlet flow specified.
        This test makes sure this version of the model is structurally sound.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 4,
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        cv = m.fs.pipeline.control_volume
        assert_units_consistent(m)

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 5
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")
        # Note that units are not consistent after time discretization.
        # This is because accumulations have different units than holdups,
        # and "delta t" factors in discretization equations have no units.

        t0 = m.fs.time.first()
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        j = next(iter(m.fs.properties.component_list))

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Fix dynamic inputs. Now these are inlet pressure and outlet
        # flow rate, as well as inlet mole fraction and temperature.
        cv.pressure[:, x0].fix()
        cv.flow_mass[:, xf].fix()
        cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix()

        # Fix initial conditions. Here, pressure and volume for all
        # non-inlet points.
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        igraph = IncidenceGraphInterface(m)
        N, M = igraph.incidence_matrix.shape
        matching = igraph.maximum_matching()
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)  # Sanity check
        self.assertEqual(len(matching), N)

    def test_dynamic_dof_forward_space(self):
        """
        Test that dynamic degrees of freedom are what we expect when using
        a forward discretization in the space domain.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
            "transformation_method": "dae.finite_difference",
            "transformation_scheme": "FORWARD",
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        cv = m.fs.pipeline.control_volume

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 2
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Inputs are inlet pressure, mole frac, and temperature, and outlet
        # flow rate.
        n_inputs = 4
        # pressure and flow mass are differential except where they are
        # specified by inputs.
        n_differential = 2 * (len(cv.length_domain) - 1)
        pred_dof = n_inputs * len(m.fs.time) + n_differential
        self.assertEqual(degrees_of_freedom(m), pred_dof)

        # Fix degrees of freedom.
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        t0 = m.fs.time.first()
        # Inputs at every point in time:
        cv.pressure[:, x0].fix()
        cv.flow_mass[:, xf].fix()
        for j in m.fs.properties.component_list:
            cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix()

        # Initial conditions:
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        igraph = IncidenceGraphInterface(m)
        N, M = igraph.incidence_matrix.shape
        matching = igraph.maximum_matching()
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)  # Sanity check
        self.assertEqual(len(matching), N)

    def test_dynamic_dof_backward_space(self):
        """
        Test that dynamic degrees of freedom are what we expect when using
        a backward discretization in the space domain.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
            "transformation_method": "dae.finite_difference",
            "transformation_scheme": "BACKWARD",
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        cv = m.fs.pipeline.control_volume

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 2
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Inputs are inlet pressure, mole frac, and temperature, and outlet
        # flow rate.
        n_inputs = 4
        # pressure and flow mass are differential except where they are
        # specified by inputs.
        n_differential = 2 * (len(cv.length_domain) - 1)
        pred_dof = n_inputs * len(m.fs.time) + n_differential
        self.assertEqual(degrees_of_freedom(m), pred_dof)

        # Fix degrees of freedom.
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        t0 = m.fs.time.first()
        # Inputs at every point in time:
        cv.pressure[:, x0].fix()
        cv.flow_mass[:, xf].fix()
        for j in m.fs.properties.component_list:
            cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix()

        # Initial conditions:
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        igraph = IncidenceGraphInterface(m)
        N, M = igraph.incidence_matrix.shape
        matching = igraph.maximum_matching()
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)  # Sanity check
        self.assertEqual(len(matching), N)

    def test_dynamic_dof_radau_space(self):
        """
        Test that dynamic degrees of freedom are what we expect when using
        a Radau discretization in the space domain.
        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
            "collocation_points": 2,
            "transformation_method": "dae.collocation",
            "transformation_scheme": "LAGRANGE-RADAU",
        }
        m.fs.pipeline = GasPipeline(**pipeline_config)

        cv = m.fs.pipeline.control_volume

        disc = pyo.TransformationFactory("dae.finite_difference")
        ntfe = 2
        disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        # Fix geometry variables
        m.fs.pipeline.diameter.fix(0.92 * pyo.units.m)
        cv.length.fix(300.0 * pyo.units.km)

        # Inputs are inlet pressure, mole frac, and temperature, and outlet
        # flow rate.
        n_inputs = 4
        # pressure and flow mass are differential except where they are
        # specified by inputs.
        n_differential = 2 * (len(cv.length_domain) - 1)
        pred_dof = n_inputs * len(m.fs.time) + n_differential
        self.assertEqual(degrees_of_freedom(m), pred_dof)

        # Fix degrees of freedom.
        x0 = cv.length_domain.first()
        xf = cv.length_domain.last()
        t0 = m.fs.time.first()
        # Inputs at every point in time:
        cv.pressure[:, x0].fix()
        cv.flow_mass[:, xf].fix()
        for j in m.fs.properties.component_list:
            cv.properties[:, x0].mole_frac_comp[j].fix()
        cv.properties[:, x0].temperature.fix()

        # Initial conditions:
        for x in cv.length_domain:
            if x != x0:
                cv.pressure[t0, x].fix()
            if x != xf:
                cv.flow_mass[t0, x].fix()

        igraph = IncidenceGraphInterface(m)
        N, M = igraph.incidence_matrix.shape
        matching = igraph.maximum_matching()
        self.assertEqual(degrees_of_freedom(m), 0)
        self.assertEqual(N, M)  # Sanity check
        self.assertEqual(len(matching), N)

    def test_dynamic_dof_legendre_space(self):
        """
        Test that trying to create a dynamic pipeline with a Legendre
        discretization throws an error. The commented code is what we
        would like to work. Debugging the Legendre discretization is not
        a high priority, so for now it is not supported.

        """
        m = pyo.ConcreteModel()
        default = {
            "dynamic": True,
            "time_set": [0.0, 20.0],
            "time_units": pyo.units.hr,
        }
        m.fs = idaes.FlowsheetBlock(**default)
        m.fs.properties = NaturalGasParameterBlock()
        pipeline_config = {
            "property_package": m.fs.properties,
            "finite_elements": 2,
            "collocation_points": 2,
            "transformation_method": "dae.collocation",
            "transformation_scheme": "LAGRANGE-LEGENDRE",
        }
        with self.assertRaisesRegex(ValueError, "Discretization scheme"):
            m.fs.pipeline = GasPipeline(**pipeline_config)

        #
        # The following is the failing test that causes Legendre
        # discretization to not be supported. If at any point we really
        # need to use a Legendre discretization, this test will need to
        # be addressed.
        #
        # cv = m.fs.pipeline.control_volume

        # disc = pyo.TransformationFactory("dae.finite_difference")
        # ntfe = 2
        # disc.apply_to(m, nfe=ntfe, wrt=m.fs.time, scheme="BACKWARD")

        ## Fix geometry variables
        # m.fs.pipeline.diameter.fix(0.92*pyo.units.m)
        # cv.length.fix(300.0*pyo.units.km)

        ## Inputs are inlet pressure, mole frac, and temperature, and outlet
        ## flow rate.
        # n_inputs = 4
        ## pressure and flow mass are differential except where they are
        ## specified by inputs.
        # n_differential = 2*(len(cv.length_domain) - 1)
        # pred_dof = n_inputs * len(m.fs.time) + n_differential
        # self.assertEqual(degrees_of_freedom(m), pred_dof)

        ## Fix degrees of freedom.
        # x0 = cv.length_domain.first()
        # xf = cv.length_domain.last()
        # t0 = m.fs.time.first()
        ## Inputs at every point in time:
        # cv.pressure[:, x0].fix()
        # cv.flow_mass[:, xf].fix()
        # for j in m.fs.properties.component_list:
        #    cv.properties[:, x0].mole_frac_comp[j].fix()
        # cv.properties[:, x0].temperature.fix()

        ## Initial conditions:
        # for x in cv.length_domain:
        #    if x != x0:
        #        cv.pressure[t0, x].fix()
        #    if x != xf:
        #        cv.flow_mass[t0, x].fix()

        # igraph = IncidenceGraphInterface(m)
        # N, M = igraph.incidence_matrix.shape
        # matching = igraph.maximum_matching()
        # self.assertEqual(degrees_of_freedom(m), 0)
        # self.assertEqual(N, M) # Sanity check
        # self.assertEqual(len(matching), N)


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest
import pandas as pd
from pyomo.common import unittest as pyo_unittest
from pyomo.opt.base.solvers import OptSolver
from idaes.apps.grid_integration.bidder import PEMParametrizedBidder
from idaes.apps.grid_integration.forecaster import PerfectForecaster
from idaes.apps.grid_integration.tests.util import (
    ExampleModel,
    ExampleForecaster,
    testing_renewable_data,
)
from idaes.apps.grid_integration.coordinator import prescient_avail
from idaes.apps.grid_integration.utils import convert_marginal_costs_to_actual_costs

day_ahead_horizon = 6
real_time_horizon = 6
# instead of using cbc, use a quasi solver to pass the _check_solver.
solver = OptSolver(type="solver")


@pytest.mark.unit
def test_creat_PEMParametrizedBidder_with_wrong_PEM_power():
    """
    This is to test when we creat the PEM bidder with a PEM power greater than the renewable power,
    there will be an error.
    """
    bidding_model_object = ExampleModel(model_data=testing_renewable_data)
    forecaster = ExampleForecaster(prediction=30)
    renewable_mw = 200
    pem_mw = 300
    pem_marginal_cost = 30
    with pytest.raises(
        ValueError, match=r".*The power of PEM is greater than the renewable power.*"
    ):
        PEM_bidder = PEMParametrizedBidder(
            bidding_model_object,
            day_ahead_horizon,
            real_time_horizon,
            solver,
            forecaster,
            renewable_mw,
            pem_marginal_cost,
            pem_mw,
        )


def wind_df():
    """
    This is to define a dataframe fed to the PerfectForecaster with example DA/RT
    capacity factors and LMPs.
    """
    start_year = 2020
    start_mon = 1
    start_day = 1
    start_date = pd.Timestamp(f"{start_year}-{start_mon:02d}-{start_day:02d} 00:00:00")
    ix = pd.date_range(
        start=start_date,
        end=start_date
        + pd.offsets.DateOffset(days=1)
        - pd.offsets.DateOffset(hours=24 - day_ahead_horizon + 1),
        freq="1H",
    )
    df = pd.DataFrame(index=ix)
    gen = testing_renewable_data.gen_name
    bus = testing_renewable_data.bus
    df[f"{gen}-DACF"] = [0.1, 0.1, 0.1, 0.8, 0.8, 0.8]
    df[f"{gen}-RTCF"] = [1.0, 0.2, 0.2, 0.75, 0.75, 0.75]
    df[f"{bus}-DALMP"] = [10] * day_ahead_horizon
    df[f"{bus}-RTLMP"] = [20] * real_time_horizon
    return df


@pytest.fixture
def bidder_object():
    """
    This is to define a bidder object.
    """
    example_wind_df = wind_df()
    forecaster = PerfectForecaster(example_wind_df)
    bidding_model_object = ExampleModel(model_data=testing_renewable_data)
    bidder_object = PEMParametrizedBidder(
        bidding_model_object=bidding_model_object,
        day_ahead_horizon=day_ahead_horizon,
        real_time_horizon=real_time_horizon,
        solver=solver,
        forecaster=forecaster,
        renewable_mw=200,
        pem_marginal_cost=30,
        pem_mw=100,
    )
    return bidder_object


@pytest.mark.component
@pytest.mark.skipif(
    not prescient_avail, reason="Prescient (optional dependency) not available"
)
def test_compute_DA_bids(bidder_object):
    """
    This is to test the if the compute_DA_bids function works correctly.
    """
    gen = bidder_object.generator
    pmin = bidder_object.bidding_model_object.pmin
    pmax = bidder_object.bidding_model_object.pmax
    pem_pmax = bidder_object.pem_mw
    pem_marginal_cost = bidder_object.pem_marginal_cost
    date = "2020-01-01"

    # test DA bidding
    bids = bidder_object.compute_day_ahead_bids(date=date, hour=0)
    expected_DA_cf = [0.1, 0.1, 0.1, 0.8, 0.8, 0.8]
    expected_bids = {}
    for t in range(day_ahead_horizon):
        expect_da_wind = expected_DA_cf[t] * pmax
        if t <= 2:
            expect_bids_curve = [(0, 0), (20, 30)]
        else:
            expect_bids_curve = [
                (0, 0),
                (60, 0),
                (160, 30),
            ]
        expect_cost_curve = convert_marginal_costs_to_actual_costs(expect_bids_curve)
        expected_bids[t] = {
            gen: {
                "p_cost": expect_cost_curve,
                "p_min": pmin,
                "p_max": expect_da_wind,
                "startup_capacity": expect_da_wind,
                "shutdown_capacity": expect_da_wind,
            }
        }
    pyo_unittest.assertStructuredAlmostEqual(first=expected_bids, second=bids)


@pytest.mark.component
@pytest.mark.skipif(
    not prescient_avail, reason="Prescient (optional dependency) not available"
)
def test_compute_RT_bids(bidder_object):
    """
    This is to test the if the compute_DA_bids function works correctly.
    """
    gen = bidder_object.generator
    pmin = bidder_object.bidding_model_object.pmin
    pmax = bidder_object.bidding_model_object.pmax
    pem_pmax = bidder_object.pem_mw
    pem_marginal_cost = bidder_object.pem_marginal_cost
    date = "2020-01-01"
    # totally 3 cases:
    # rt_wind - realized_day_ahead_dispatches > P_pem_max (t = 0)
    # 0 <= rt_wind - realized_day_ahead_dispatches <= P_pem_max (t = 1, 2, 3, 4)
    # rt_wind <= realized_day_ahead_dispatches (t = 5)
    realized_day_ahead_dispatches = [20, 20, 20, 120, 120, 180]
    realized_day_ahead_prices = None

    # testing RT bidding
    bids = bidder_object.compute_real_time_bids(
        date=date,
        hour=0,
        realized_day_ahead_dispatches=realized_day_ahead_dispatches,
        realized_day_ahead_prices=realized_day_ahead_prices,
    )

    expected_RT_cf = [1.0, 0.2, 0.2, 0.75, 0.75, 0.75]
    expected_bids = {}
    for t in range(real_time_horizon):
        expect_rt_wind = expected_RT_cf[t] * pmax
        if t == 0:
            expect_bids_curve = [
                (0, 0),
                (80, 0),
                (180, 30),
            ]
        elif t in [1, 2]:
            expect_bids_curve = [(0, 0), (20, 30)]
        elif t in [3, 4]:
            expect_bids_curve = [(0, 0), (30, 30)]
        else:
            expect_bids_curve = [(0, 0)]
        expect_cost_curve = convert_marginal_costs_to_actual_costs(expect_bids_curve)
        expected_bids[t] = {
            gen: {
                "p_cost": expect_cost_curve,
                "p_min": pmin,
                "p_max": max([p[0] for p in expect_cost_curve]),
                "startup_capacity": expect_rt_wind,
                "shutdown_capacity": expect_rt_wind,
            }
        }
    pyo_unittest.assertStructuredAlmostEqual(first=expected_bids, second=bids)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Case B31A - Natural Gas Combined Cycle (NGCC) plant_gross_power
Reference: NETL-PUB-22638
Cost and Performance Baseline for Fossil Energy Plants Volume 1:
Bituminous Coal and Natural Gas to Electricity; https://doi.org/10.2172/1569246
Author: A. Deshpande and M. Zamarripa
"""
import pytest
from idaes.models_extra.power_generation.costing.power_plant_capcost import (
    QGESSCosting,
    QGESSCostingData,
)
from idaes.core.util.model_statistics import degrees_of_freedom
import pyomo.environ as pyo
from idaes.core import FlowsheetBlock, UnitModelBlock, UnitModelCostingBlock
from idaes.core.solvers import get_solver
from pyomo.environ import units as pyunits
from pyomo.util.check_units import assert_units_consistent

# Get default solver for testing
solver = get_solver()


@pytest.fixture(scope="module")
def build_costing():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.costing = QGESSCosting()
    CE_index_year = "2018"

    # Accounts with Feedwater Flow to HP section of HRSG, as the
    # reference/scaling parameter - Exhibit 5-15
    FW_accounts = ["3.1", "3.3", "8.4"]

    m.fs.b1 = UnitModelBlock()
    m.fs.b1.feedwater_flowrate = pyo.Var(
        initialize=1085751, units=pyunits.lb / pyunits.hr
    )
    m.fs.b1.feedwater_flowrate.fix()

    m.fs.b1.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": FW_accounts,
            "scaled_param": m.fs.b1.feedwater_flowrate,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    return m


@pytest.mark.unit
def test_get_costing(build_costing):
    m = build_costing

    assert isinstance(m.fs.b1.costing.total_plant_cost, pyo.Var)
    assert hasattr(m.fs.b1.costing, "bare_erected_cost")
    assert isinstance(m.fs.b1.costing.total_plant_cost_eq, pyo.Constraint)
    assert degrees_of_freedom(m) == 0


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_units1_costing(build_costing):
    m = build_costing
    CE_index_year = "2018"

    # Accounts with Feedwater Flow to HP section of HRSG, as the
    # reference/scaling parameter - Exhibit 5-15
    FW_accounts = ["3.1", "3.3", "8.4"]

    # Accounts with Raw water withdrawal as the reference/scaling parameter
    # Exhibit 5-14
    RW_withdraw_accounts = ["3.2", "3.4", "3.5", "9.5", "14.6"]
    m.fs.b2 = UnitModelBlock()

    m.fs.b2.raw_water_withdrawal = pyo.Var(
        initialize=2902, units=pyunits.gal / pyunits.min
    )
    m.fs.b2.raw_water_withdrawal.fix()
    m.fs.b2.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": RW_withdraw_accounts,
            "scaled_param": m.fs.b2.raw_water_withdrawal,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with fuel gas flowrate as the reference/scaling parameter
    # Exhibit 5-15 stream 2, Exhibit 5-8
    FuelG_accounts = ["3.6", "3.9", "6.1", "6.3", "6.4"]
    m.fs.b3 = UnitModelBlock()
    # Obtain Fuel gas flowrate in acm
    fuelgas_value = 205630  # lb/hr

    m.fs.b3.fg_flowrate = pyo.Var(
        initialize=fuelgas_value, units=pyunits.lb / pyunits.hr
    )
    m.fs.b3.fg_flowrate.fix()
    m.fs.b3.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": FuelG_accounts,
            "scaled_param": m.fs.b3.fg_flowrate,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with process water discharge as the reference/scaling parameter
    # Exhibit 5-14
    PW_discharge_accounts = ["3.7"]
    m.fs.b4 = UnitModelBlock()

    m.fs.b4.process_water_discharge = pyo.Var(
        initialize=657, units=pyunits.gal / pyunits.min
    )
    m.fs.b4.process_water_discharge.fix()
    m.fs.b4.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": PW_discharge_accounts,
            "scaled_param": m.fs.b4.process_water_discharge,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Initialize costing
    QGESSCostingData.costing_initialization(m.fs.costing)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)

    # check unit consistency
    assert_units_consistent(m)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # Accounts with raw water withdrawal as reference parameter
    assert pytest.approx(26.435, abs=0.5) == sum(
        pyo.value(m.fs.b2.costing.total_plant_cost[ac]) for ac in RW_withdraw_accounts
    )
    # Accounts with fuel gas as reference parameter
    assert pytest.approx(158.415, abs=0.5) == sum(
        pyo.value(m.fs.b3.costing.total_plant_cost[ac]) for ac in FuelG_accounts
    )

    # Accounts with process water discharge as reference parameter
    assert pytest.approx(11.608, abs=0.5) == sum(
        pyo.value(m.fs.b4.costing.total_plant_cost[ac]) for ac in PW_discharge_accounts
    )


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_units2_costing(build_costing):
    m = build_costing
    CE_index_year = "2018"
    # Accounts with flue gas flowrate as the reference/scaling parameter
    # Exhibit 5-15 stream 3, Exhibit 5-8
    FG_accounts = ["7.6"]
    m.fs.b5 = UnitModelBlock()
    # Obtain Flue gas flowrate in acm
    fluegas_value = (8658430 / 60) / 0.025  # ft3/min

    m.fs.b5.fg_flowrate = pyo.Var(
        initialize=fluegas_value, units=pyunits.ft**3 / pyunits.min
    )
    m.fs.b5.fg_flowrate.fix()
    m.fs.b5.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": FG_accounts,
            "scaled_param": m.fs.b5.fg_flowrate,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with combustion turbine gross power as the reference/scaling
    # parameter
    # Exhibit 5-9
    CT_grosspower_accounts = ["6.5"]
    m.fs.b6 = UnitModelBlock()
    # Obtain combustion turbine gross power in kW
    CT_gross_power = 477 * 1000  # kW

    m.fs.b6.ct_gross_power = pyo.Var(initialize=CT_gross_power, units=pyunits.kW)
    m.fs.b6.ct_gross_power.fix()
    m.fs.b6.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": CT_grosspower_accounts,
            "scaled_param": m.fs.b6.ct_gross_power,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with HRSG duty as the reference/scaling parameter
    # Exhibit 5-8, streams 3 and 4
    HRSG_duty_accounts = ["7.1", "7.2"]
    m.fs.b7 = UnitModelBlock()
    # Obtain HRSG duty in MMBtu/hr, overall energy balance
    HRSG_duty = -(-538.1 + 277.1) * 8658430 / (10**6)  # MMBtu/hr

    m.fs.b7.hrsg_duty = pyo.Var(initialize=HRSG_duty, units=pyunits.MBtu / pyunits.hr)
    m.fs.b7.hrsg_duty.fix()
    m.fs.b7.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": HRSG_duty_accounts,
            "scaled_param": m.fs.b7.hrsg_duty,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with gas flow to stack as the reference/scaling parameter
    # Exhibit 5-8, stream 4
    Stack_flow_gas_accounts = ["7.3", "7.4", "7.5"]
    m.fs.b8 = UnitModelBlock()
    # Obtain gas flowrate to stack in ft3/min
    stack_flow_gas = (8658430 / 60) / 0.061  # ft3/min

    m.fs.b8.stack_flow_gas = pyo.Var(
        initialize=stack_flow_gas, units=pyunits.ft**3 / pyunits.min
    )
    m.fs.b8.stack_flow_gas.fix()
    m.fs.b8.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": Stack_flow_gas_accounts,
            "scaled_param": m.fs.b8.stack_flow_gas,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Initialize costing
    QGESSCostingData.costing_initialization(m.fs.costing)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)

    # check unit consistency
    assert_units_consistent(m)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok
    # Accounts with HRSG duty as reference parameter
    assert pytest.approx(90.794, abs=0.1) == sum(
        pyo.value(m.fs.b7.costing.total_plant_cost[ac]) for ac in HRSG_duty_accounts
    )


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_units3_costing(build_costing):
    m = build_costing
    CE_index_year = "2018"
    # Accounts with steam turbine gross power as the reference/scaling
    # parameter
    # Exhibit 5-9
    Steam_turbine_gross_power_accounts = ["8.1", "8.2", "8.5", "14.3"]
    m.fs.b9 = UnitModelBlock()
    # Obtain steam turbine gross power in kW
    ST_gross_power = 263 * 1000  # kW

    m.fs.b9.st_gross_power = pyo.Var(initialize=ST_gross_power, units=pyunits.kW)
    m.fs.b9.st_gross_power.fix()
    m.fs.b9.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": Steam_turbine_gross_power_accounts,
            "scaled_param": m.fs.b9.st_gross_power,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with condenser duty as the reference/scaling parameter
    # Exhibit 5-9
    Condenser_duty_accounts = ["8.3"]
    m.fs.b10 = UnitModelBlock()
    # Obtain condenser duty in MMBtu/hr
    condenser_duty = 1332  # MMBtu/hr

    m.fs.b10.cond_duty = pyo.Var(
        initialize=condenser_duty, units=pyunits.MBtu / pyunits.hr
    )
    m.fs.b10.cond_duty.fix()
    m.fs.b10.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": Condenser_duty_accounts,
            "scaled_param": m.fs.b10.cond_duty,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with cooling tower duty as the reference/scaling parameter
    # Exhibit 5-16
    Cooling_tower_accounts = ["9.1"]
    m.fs.b11 = UnitModelBlock()
    # Obtain cooling tower duty in MMBtu/hr (includes condenser, Acid gas
    # removal, and other cooling loads)
    cooling_tower_duty = 1357  # MMBtu/hr

    m.fs.b11.cool_tower_duty = pyo.Var(
        initialize=cooling_tower_duty, units=pyunits.MBtu / pyunits.hr
    )
    m.fs.b11.cool_tower_duty.fix()
    m.fs.b11.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": Cooling_tower_accounts,
            "scaled_param": m.fs.b11.cool_tower_duty,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with circulating water flowrate as the reference/scaling
    # parameter
    Circ_water_accounts = ["9.2", "9.3", "9.4", "9.6", "9.7", "14.5"]
    m.fs.b12 = UnitModelBlock()
    # Obtain circulating water flowrate in gal/min
    cir_water_flowrate = 217555  # gal/min

    m.fs.b12.circ_water_flow = pyo.Var(
        initialize=cir_water_flowrate, units=pyunits.gal / pyunits.min
    )
    m.fs.b12.circ_water_flow.fix()
    m.fs.b12.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": Circ_water_accounts,
            "scaled_param": m.fs.b12.circ_water_flow,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with total plant gross power as the reference/scaling parameter
    # Exhibit 5-9
    plant_gross_power_accounts = [
        "11.1",
        "11.7",
        "11.9",
        "13.1",
        "13.2",
        "13.3",
        "14.4",
        "14.7",
        "14.8",
        "14.9",
        "14.10",
    ]
    m.fs.b13 = UnitModelBlock()
    # Obtain total plant gross power in kW
    plant_gross_power = 740000  # kW

    m.fs.b13.gross_power = pyo.Var(initialize=plant_gross_power, units=pyunits.kW)
    m.fs.b13.gross_power.fix()
    m.fs.b13.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": plant_gross_power_accounts,
            "scaled_param": m.fs.b13.gross_power,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with auxiliary load as the reference/scaling parameter
    # Exhibit 5-9
    auxiliary_load_accounts = [
        "11.2",
        "11.3",
        "11.4",
        "11.5",
        "11.6",
        "12.1",
        "12.2",
        "12.3",
        "12.4",
        "12.5",
        "12.6",
        "12.7",
        "12.8",
        "12.9",
    ]
    m.fs.b14 = UnitModelBlock()
    # Obtain auxiliary load in kW
    aux_load = 14 * 1000  # kW

    m.fs.b14.auxiliary_load = pyo.Var(initialize=aux_load, units=pyunits.kW)
    m.fs.b14.auxiliary_load.fix()
    m.fs.b14.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": auxiliary_load_accounts,
            "scaled_param": m.fs.b14.auxiliary_load,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with STG,CTG output as the reference/scaling parameter
    # Case B31A Account 9 - Pg 501 rev 4 baseline report
    stg_ctg_accounts = ["11.8"]
    m.fs.b15 = UnitModelBlock()
    # Obtain STG,CTG output in kW
    stg_ctg_op = 689800  # kW

    m.fs.b15.stg_ctg_output = pyo.Var(initialize=stg_ctg_op, units=pyunits.kW)
    m.fs.b15.stg_ctg_output.fix()
    m.fs.b15.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": stg_ctg_accounts,
            "scaled_param": m.fs.b15.stg_ctg_output,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Accounts with gas turbine power as the reference/scaling parameter
    # Exhibit 5-9
    gasturbine_accounts = ["14.1"]
    m.fs.b16 = UnitModelBlock()
    # Obtain gas turbine power in kW
    gt_power = 477 * 1000  # kW

    m.fs.b16.gas_turbine_power = pyo.Var(initialize=gt_power, units=pyunits.kW)
    m.fs.b16.gas_turbine_power.fix()
    m.fs.b16.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=QGESSCostingData.get_PP_costing,
        costing_method_arguments={
            "cost_accounts": gasturbine_accounts,
            "scaled_param": m.fs.b16.gas_turbine_power,
            "tech": 6,
            "ccs": "B",
            "CE_index_year": CE_index_year,
        },
    )

    # Initialize costing
    QGESSCostingData.costing_initialization(m.fs.costing)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)

    # check unit consistency
    assert_units_consistent(m)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # Accounts with condenser duty as reference parameter
    assert pytest.approx(14.27, abs=0.1) == sum(
        pyo.value(m.fs.b10.costing.total_plant_cost[ac])
        for ac in Condenser_duty_accounts
    )

    # Accounts with cooling tower duty as reference parameter
    assert pytest.approx(14.73, abs=0.2) == sum(
        pyo.value(m.fs.b11.costing.total_plant_cost[ac])
        for ac in Cooling_tower_accounts
    )


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_flowsheet_costing(build_costing):
    m = build_costing
    # Build cost constraints
    m.fs.costing.build_process_costs(
        net_power=None,
        fixed_OM=False,
        variable_OM=False,
        resources=None,
        rates=None,
        prices=None,
        fuel="natural_gas",
    )

    # Initialize costing
    QGESSCostingData.costing_initialization(m.fs.costing)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)

    # check unit consistency
    assert_units_consistent(m)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # Verify total plant costs
    assert pytest.approx(574.09, abs=0.1) == pyo.value(m.fs.costing.total_TPC)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import numpy as np
from math import sqrt
from idaes.apps.matopt.materials import (
    Atom,
    Canvas,
    Design,
    LinearTiling,
    PlanarTiling,
    CubicTiling,
)
from idaes.apps.matopt.materials.geometry import (
    Shape,
    Cuboctahedron,
    Parallelepiped,
    Rhombohedron,
    RectPrism,
    Cube,
    Cylinder,
    CylindricalSector,
)
from idaes.apps.matopt.materials.lattices import (
    FCCLattice,
    CubicLattice,
    PerovskiteLattice,
    DiamondLattice,
    WurtziteLattice,
)
from idaes.apps.matopt.materials.transform_func import (
    ShiftFunc,
    ScaleFunc,
    RotateFunc,
    ReflectFunc,
)
from idaes.apps.matopt.opt import (
    Coef,
    LinearExpr,
    MatOptModel,
    SumNeighborSites,
    SumNeighborBonds,
    SumSites,
    SumBonds,
    SumSiteTypes,
    SumBondTypes,
    SumSitesAndTypes,
    SumBondsAndTypes,
    SumConfs,
    SumSitesAndConfs,
    LessThan,
    EqualTo,
    GreaterThan,
    FixedTo,
    Disallow,
    PiecewiseLinear,
    Implies,
    NegImplies,
    ImpliesSiteCombination,
    ImpliesNeighbors,
)
from idaes.apps.matopt.util.util import areEqual
import pytest
from test_matopt_objects_construction import *


@pytest.mark.unit
def test_functionality_FCCLattice():
    lattice = test_construct_FCCLattice()
    assert lattice.areNeighbors(np.zeros(3, dtype=float), np.array([0.0, -0.5, 0.5]))
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert areEqual(lattice.IAD, sqrt(2) / 2, 1e-4)
    assert areEqual(lattice.FCC111LayerSpacing, 1 / sqrt(3), 1e-4)
    assert areEqual(lattice.FCC100LayerSpacing, 1 / 2, 1e-4)
    assert areEqual(lattice.FCC110LayerSpacing, sqrt(2) / 2, 1e-4)


@pytest.mark.unit
def test_functionality_CubicLattice():
    lattice = test_construct_CubicLattice()
    assert lattice.areNeighbors(np.zeros(3, dtype=float), np.array([1.0, 0.0, 0.0]))
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert areEqual(lattice.IAD, 1.0, 1e-4)


@pytest.mark.unit
def test_functionality_DiamondLattice():
    lattice = test_construct_DiamondLattice()
    assert lattice.areNeighbors(np.zeros(3, dtype=float), np.array([0.25, 0.25, 0.25]))
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert lattice.isASite(np.zeros(3, dtype=float))
    assert lattice.isBSite(np.array([0.25, 0.25, 0.25]))
    assert areEqual(lattice.IAD, sqrt(3) / 4, 1e-4)
    assert areEqual(lattice.getLayerSpacing("100"), 1 / 4, 1e-4)
    assert areEqual(lattice.getLayerSpacing("110"), sqrt(2) / 4, 1e-4)
    assert areEqual(lattice.getLayerSpacing("111"), sqrt(3) / 3, 1e-4)
    assert areEqual(lattice.getLayerSpacing("112"), sqrt(6) / 12, 1e-4)
    assert areEqual(lattice.getShellSpacing("100"), sqrt(8) / 4, 1e-4)
    assert areEqual(lattice.getShellSpacing("112"), sqrt(2) / 4, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount("100"), 4, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount("110"), 2, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount("111"), 3, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount("112"), 6, 1e-4)


@pytest.mark.unit
def test_functionality_WurtziteLattice():
    lattice = test_construct_WurtziteLattice()
    assert lattice.areNeighbors(
        np.zeros(3, dtype=float), np.array([0.0, 0.0, -3 / sqrt(24)])
    )
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert lattice.isASite(np.zeros(3, dtype=float))
    assert lattice.isBSite(np.array([0.0, 0.0, -3 / sqrt(24)]))
    assert areEqual(lattice.IAD, sqrt(3 / 8), 1e-4)
    assert areEqual(lattice.getLayerSpacing("0001"), sqrt(2 / 3), 1e-4)
    assert areEqual(lattice.getLayerSpacing("1100"), sqrt(3) / 2, 1e-4)
    assert areEqual(lattice.getLayerSpacing("1120"), 1, 1e-4)
    assert areEqual(lattice.getShellSpacing("0001"), sqrt(3) / 2, 1e-4)
    assert areEqual(lattice.getShellSpacing("1100"), 1, 1e-4)
    assert areEqual(lattice.getShellSpacing("1120"), 1, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount("0001"), 2, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount("1100"), 2, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount("1120"), 1, 1e-4)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES 0D Fixed Bed Reactor model.
"""

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Var,
    value,
    check_optimal_termination,
    units as pyunits,
)
from pyomo.dae import DerivativeVar
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    EnergyBalanceType,
    UnitModelBlockData,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.constants import Constants as constants
import idaes.logger as idaeslog
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

__author__ = "Chinedu Okoli, Andrew Lee"


# Assumptions:
# Only Vap and Sol phases, which are explicitly named as such
# Perfect mixing in Vap phase
# Static solid phase


@declare_process_block_class("FixedBed0D")
class FixedBed0DData(UnitModelBlockData):
    """
    0D Fixed Bed Reactor Model Class
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([True]),
            default=True,
            description="Dynamic model flag - must be True",
            doc="""Indicates whether this model will be dynamic or not,
**default** = True. Fixed beds must be dynamic.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=True,
            domain=In([True]),
            description="Holdup construction flag - must be True",
            doc="""Indicates whether holdup terms should be constructed or not.
**default** - True. Fixed bed reactors must be dynamic, thus this must be
True.""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "gas_property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for gas phase",
            doc="""Property parameter object used to define property calculations
for the gas phase, **default** - None.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "gas_property_package_args",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Arguments to use for constructing gas phase "
            "property packages",
            doc="""A ConfigBlock with arguments to be passed to a gas phase
property block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "solid_property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for solid phase",
            doc="""Property parameter object used to define property calculations
for the solid phase, **default** - None.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "solid_property_package_args",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Arguments to use for constructing solid phase "
            "property packages",
            doc="""A ConfigBlock with arguments to be passed to a solid phase
property block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "reaction_package",
        ConfigValue(
            default=None,
            # Removed domain argument due to limitations in ReactionBase for
            # heterogeneous systems
            description="Reaction package to use for unit",
            doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "reaction_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}""",
        ),
    )

    def build(self):
        # Call UnitModel.build to setup dynamics
        super().build()

        # Get units meta data from property packages (only solid needed)
        units_meta_solid = (
            self.config.solid_property_package.get_metadata().get_derived_units
        )

        # Build Gas Phase StateBlock
        # This block only needed so gas conc. to rxn block is calculated
        # Defined state is set to True so that the "sum(mole_frac)=1" eqn in
        # the gas state block is deactivated.
        self.gas = self.config.gas_property_package.state_block_class(
            self.flowsheet().time,
            parameters=self.config.gas_property_package,
            defined_state=True,
        )

        # Build Solid Phase StateBlock
        # As there is no solid flow, only need a single state block
        # Defined state is set to True so that the "sum(mass_frac)=1" eqn in
        # the solid state block is deactivated. This is done here as there is
        # currently no way to deactivate the constraint at the initial time
        # for batch systems (i.e. no inlet or outlet ports).
        # The "sum(mass_frac)=1 for all t neq 0" eqn is written in the
        # unit model instead
        self.solids = self.config.solid_property_package.state_block_class(
            self.flowsheet().time,
            parameters=self.config.solid_property_package,
            defined_state=True,
        )

        tmp_dict = dict(**self.config.reaction_package_args)
        tmp_dict["gas_state_block"] = self.gas
        tmp_dict["solid_state_block"] = self.solids
        tmp_dict["parameters"] = self.config.reaction_package
        self.reactions = self.config.reaction_package.reaction_block_class(
            self.flowsheet().time,
            doc="Reaction properties in control volume",
            **tmp_dict,
        )

        # Volume of reactor
        self.bed_diameter = Var(
            initialize=1, doc="Reactor diameter", units=units_meta_solid("length")
        )
        self.bed_height = Var(
            initialize=1, doc="Bed length", units=units_meta_solid("length")
        )
        self.volume_bed = Var(
            initialize=1.0,
            doc="Volume of the reactor bed",
            units=units_meta_solid("volume"),
        )

        @self.Constraint(doc="Calculating volume of the reactor bed")
        def volume_bed_constraint(b):
            return b.volume_bed == (
                constants.pi * b.bed_height * (0.5 * b.bed_diameter) ** 2
            )

        # Solid phase material balance
        # Volume of solid
        self.volume_solid = Var(
            self.flowsheet().config.time,
            initialize=1.0,
            doc="Solids phase volume including particle pores",
            units=units_meta_solid("volume"),
        )

        @self.Constraint(
            self.flowsheet().config.time, doc="Calculating solid phase volume"
        )
        def volume_solid_constraint(b, t):
            return b.volume_solid[t] == (
                b.volume_bed * (1 - b.solids[t].params.voidage)
            )

        # Accumulation equal to mass transfer/reaction
        self.solids_material_holdup = Var(
            self.flowsheet().time,
            self.config.solid_property_package.component_list,
            initialize=1,
            doc="Solid phase component holdups",
            units=units_meta_solid("mass"),
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.solid_property_package.component_list,
            doc="Solid phase material holdup constraints",
        )
        def solids_material_holdup_constraints(b, t, j):
            return b.solids_material_holdup[t, j] == (
                b.volume_solid[t] * b.solids[t].get_material_density_terms("Sol", j)
            )

        self.solids_material_accumulation = DerivativeVar(
            self.solids_material_holdup,
            wrt=self.flowsheet().config.time,
            doc="Solids material accumulation",
            units=units_meta_solid("mass"),
        )
        # should be mass/time, to get around DAE limitations var is
        # divided by time units when used

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.solid_property_package.component_list,
            doc="Solid phase material accumulation constraints",
        )
        def solids_material_accumulation_constraints(b, t, j):
            return b.solids_material_accumulation[t, j] / units_meta_solid(
                "time"
            ) == b.volume_solid[t] * b.solids[t].params.mw_comp[j] * sum(
                b.reactions[t].reaction_rate[r]
                * b.config.reaction_package.rate_reaction_stoichiometry[r, "Sol", j]
                for r in b.config.reaction_package.rate_reaction_idx
            )

        # Add solid mass variable and constraint for TGA tracking
        self.mass_solids = Var(
            self.flowsheet().config.time,
            doc="Total mass of solids",
            units=units_meta_solid("mass"),
        )

        @self.Constraint(
            self.flowsheet().config.time, doc="Calculating total mass of solids"
        )
        def mass_solids_constraint(b, t):
            return b.mass_solids[t] == b.volume_solid[t] * sum(
                b.solids[t].get_material_density_terms("Sol", j)
                for j in b.config.solid_property_package.component_list
            )

        # Sum of mass fractions at all time equals 1
        @self.Constraint(
            self.flowsheet().config.time, doc="Sum of mass fractions at all time"
        )
        def sum_component_constraint(b, t):
            if t == b.flowsheet().config.time.first():
                return Constraint.Skip
            else:
                return 1 == sum(
                    b.solids[t].mass_frac_comp[j]
                    for j in b.solids[t].params.component_list
                )

        if self.config.energy_balance_type != EnergyBalanceType.none:
            # Solid phase energy balance
            # Accumulation equal to heat transfer
            self.solids_energy_holdup = Var(
                self.flowsheet().time,
                initialize=1,
                doc="Solid phase energy holdup",
                units=units_meta_solid("energy"),
            )

            @self.Constraint(
                self.flowsheet().config.time,
                doc="Solid phase energy holdup constraints",
            )
            def solids_energy_holdup_constraints(b, t):
                return b.solids_energy_holdup[t] == (
                    self.volume_solid[t]
                    * pyunits.convert(
                        b.solids[t].get_energy_density_terms("Sol"),
                        to_units=units_meta_solid("energy")
                        / units_meta_solid("volume"),
                    )
                )

            self.solids_energy_accumulation = DerivativeVar(
                self.solids_energy_holdup,
                wrt=self.flowsheet().config.time,
                doc="Solids energy accumulation",
                units=units_meta_solid("energy"),
            )
            # should be energy/time, to get around DAE limitations var is
            # divided by time units when used

            @self.Constraint(
                self.flowsheet().config.time,
                doc="Solid phase energy accumulation constraints",
            )
            def solids_energy_accumulation_constraints(b, t):
                return b.solids_energy_accumulation[t] / units_meta_solid(
                    "time"
                ) == -sum(
                    b.reactions[t].reaction_rate[r]
                    * b.volume_solid[t]
                    * pyunits.convert(
                        b.reactions[t].dh_rxn[r],
                        to_units=units_meta_solid("energy_mole"),
                    )
                    for r in b.config.reaction_package.rate_reaction_idx
                )

        if self.config.energy_balance_type == EnergyBalanceType.none:
            # Fix solids temperature to initial value for isothermal conditions
            @self.Constraint(
                self.flowsheet().config.time, doc="Isothermal solid phase constraint"
            )
            def isothermal_solid_phase(b, t):
                if t == b.flowsheet().config.time.first():
                    return Constraint.Skip
                else:
                    return b.solids[t].temperature == b.solids[0].temperature

    def initialize_build(
        blk,
        gas_phase_state_args=None,
        solid_phase_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for FB0D unit.

        Keyword Arguments:
            gas_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            solid_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """

        # Set solver options
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")
        opt = get_solver(solver, optarg)  # create solver
        opt.options = optarg

        # ---------------------------------------------------------------------
        # Initialize control volume block
        init_log.info("Initialize Thermophysical Properties")

        # Initialize gas_phase block
        gas_phase_flags = blk.gas.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=gas_phase_state_args
        )

        # Initialize solid_phase properties block
        solid_phase_flags = blk.solids.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=solid_phase_state_args,
        )

        init_log.info("Initialize reaction properties")
        # Initialize reactions
        blk.reactions.initialize(outlvl=outlvl, optarg=optarg, solver=solver)

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, outlvl) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.gas.release_state(gas_phase_flags, outlvl)
        blk.solids.release_state(solid_phase_flags, outlvl)

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{blk.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        init_log.info(f"Initialization Complete: {idaeslog.condition(res)}")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # scale some variables
        if hasattr(self, "bed_diameter"):
            if iscale.get_scaling_factor(self.bed_diameter) is None:
                sf = 1 / value(self.bed_diameter)
                iscale.set_scaling_factor(self.bed_diameter, 1e-1 * sf)

        if hasattr(self, "bed_height"):
            if iscale.get_scaling_factor(self.bed_height) is None:
                sf = 1 / value(self.bed_height)
                iscale.set_scaling_factor(self.bed_height, 1e-1 * sf)

        if hasattr(self, "volume_bed"):
            if iscale.get_scaling_factor(self.volume_bed) is None:
                sf = 1 / value(
                    constants.pi * self.bed_height * (0.5 * self.bed_diameter) ** 2
                )
                iscale.set_scaling_factor(self.volume_bed, 1e-1 * sf)

        # Scale some constraints
        if hasattr(self, "volume_bed_constraint"):
            for c in self.volume_bed_constraint.values():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.volume_bed), overwrite=False
                )

        if hasattr(self, "volume_solid_constraint"):
            for t, c in self.volume_solid_constraint.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.volume_bed), overwrite=False
                )

        if hasattr(self, "solids_material_holdup_constraints"):
            for (t, j), c in self.solids_material_holdup_constraints.items():
                sf1 = iscale.get_scaling_factor(self.volume_bed)
                sf2 = iscale.get_scaling_factor(self.solids[t].dens_mass_skeletal)
                iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if hasattr(self, "solids_material_accumulation_constraints"):
            for (t, j), c in self.solids_material_accumulation_constraints.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.volume_bed), overwrite=False
                )

        if hasattr(self, "mass_solids_constraint"):
            for t, c in self.mass_solids_constraint.items():
                sf1 = iscale.get_scaling_factor(self.volume_bed)
                sf2 = iscale.get_scaling_factor(self.solids[t].dens_mass_skeletal)
                iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if hasattr(self, "sum_component_constraint"):
            # Get a single representative value in component list
            for j in self.config.solid_property_package.component_list:
                break
            for t, c in self.sum_component_constraint.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.solids[t].mass_frac_comp[j]),
                    overwrite=False,
                )

        if self.config.energy_balance_type != EnergyBalanceType.none:

            if hasattr(self, "solids_energy_holdup_constraints"):
                for t, c in self.solids_energy_holdup_constraints.items():
                    sf1 = iscale.get_scaling_factor(self.volume_bed)
                    sf2 = iscale.get_scaling_factor(self.solids[t].enth_mass)
                    iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

            if hasattr(self, "solids_energy_accumulation_constraints"):
                for t, c in self.solids_energy_accumulation_constraints.items():
                    sf1 = iscale.get_scaling_factor(self.volume_bed)
                    sf2 = iscale.get_scaling_factor(self.solids[t].enth_mass)
                    iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if self.config.energy_balance_type == EnergyBalanceType.none:

            if hasattr(self, "isothermal_solid_phase"):
                for t, c in self.isothermal_solid_phase.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.solids[t].temperature),
                        overwrite=False,
                    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Convenience plotting functions for time-dependent variables.
"""

__author__ = "John Eslick"

import matplotlib.pyplot as plt

import pyomo.environ as pyo
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice


def dynamic_value_list(*args, units=None):
    """
    Create a list of values from a Pyomo component. Multiple models representing
    different time periods can also be combined by providing multiple arguments
    to assemble into one time series.

    Args:
        Positional arguments (): Multiple Pyomo components indexed by time, or
            time sets. Slices are ok.

    Returns:
        (list) with the time indexed Pyomo component values concatenated for
            plotting
    """
    l = []
    for v in args:
        if isinstance(v, pyo.Set):
            l += [t for t in v]
        elif isinstance(v, IndexedComponent_slice):
            if units is None:
                l += [pyo.value(sv) for sv in v]
            else:
                l += [pyo.value(pyo.units.convert(sv, units)) for sv in v]
        else:
            if units is None:
                l += [pyo.value(v[t]) for t in v]
            else:
                l += [pyo.value(pyo.units.convert(v[t], units)) for t in v]
    return l


def plot_grid(
    x, y, xlabel, ylabel, cols=1, rows=1, same_x=False, ylabel_title=True, to_file=None
):
    """Make a grid of plots.

    Args:
        x: values for x-axis. If all plots have the same x-axis the same_x arg
            should be True, and only one x-value list is need, otherwise x should
            be a list of x value lists.
        y: list of lists of y values for plots
        xlabel: if same_x=True, a x-axis label, if same_x=False, list of xaxis
            labels
        ylable: list of y-axis labels
        cols: number of columns in plot grid
        rows: number of rows in plot grid
        same_x: If True use the same x-axis for all plots, otherwise expect a
            list of x-axis data for the x-axis of each plot
        ylabel_title: If True, use the ylabel for the plot title, otherwise label
            the y-axis

    Returns:
        matplotlib.pyplot
    """
    assert cols * rows >= len(y)
    for i, yi in enumerate(y):
        yilabel = ylabel[i]
        if same_x:
            xi = x
            xilabel = xlabel
        else:
            xi = x[i]
            xilabel = xlabel[i]

        plt.subplot(rows, cols, i + 1)
        plt.xlabel(xilabel)
        plt.plot(xi, yi)
        if ylabel_title:
            plt.title(yilabel)
        else:
            plt.ylabel(yilabel)

    plt.tight_layout()
    if to_file:
        plt.savefig(to_file)
    else:
        plt.show()
    return plt


def plot_grid_dynamic(
    x, y, xlabel, ylabel, yunits=None, cols=1, rows=1, ylabel_title=True, to_file=None
):
    """Make a grid of dynamic plots, where the x-axices are time.

    Args:
        x: time set or list of time values
        y: list of lists of y values for plots.  The y value lists can be
            time-indexed pyo components or time slices
        xlabel: time axis label
        ylable: list of y-axis labels
        yunits: list of y-axis units, if units are provided, the y-values will
            be converted to the requested units for plotting
        cols: number of columns in plot grid
        rows: number of rows in plot grid
        ylabel_title: If True, use the ylabel for the plot title, otherwise label
            the y-axis

    Returns:
        matplotlib.pyplot
    """
    y_list = []
    if isinstance(x, pyo.Component):
        time = dynamic_value_list(x)
    else:
        time = x
    for i, yi in enumerate(y):
        if isinstance(yi, (pyo.Var, pyo.Expression, IndexedComponent_slice)):
            if yunits is not None:
                y_list.append(dynamic_value_list(yi, units=yunits[i]))
            else:
                y_list.append(dynamic_value_list(yi))
        else:
            y_list.append(yi)
    return plot_grid(
        x=time,
        y=y_list,
        xlabel=xlabel,
        ylabel=ylabel,
        cols=cols,
        rows=rows,
        same_x=True,
        ylabel_title=ylabel_title,
        to_file=to_file,
    )


def plot_dynamic(time, y, ylabel, xlabel="time (s)", title=None, legend=None):
    """
    Plot time dependent variables with pyplot.

    Args:
        time (ContinuousSet or list-like): Time index set
        y (list-like of list-likes of Var, Expression, Reference, or float):
            List of quantities to plot (multiple quantities can be plotted). Each
            quantity in the list should be indexed only by time. If you want to
            plot something that is not indexed only by time, you can create a
            Pyomo Reference with the correct indexing.
        ylabel (str): Y-axis label, required
        xlabel (str): X-axis label, default = 'time (s)'
        title (str or None): Plot title, default = None
        legend (list-like of str or None): Legend string for each y,
            default = None

    Returns:
        None
    """
    y2 = [None] * len(y)
    for i, z in enumerate(y):
        if isinstance(z, (list, tuple)):
            y2[i] = y[i]
        else:
            y2[i] = [pyo.value(z[t]) for t in time]
    for q in y2:
        plt.plot(time, q)
    if legend is not None:
        plt.legend(legend)
    if title is not None:
        plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.tight_layout()
    plt.show()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for CLC heterogeneous reaction block; tests for construction and solve
Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Var,
    Constraint,
    value,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentMap
from pyomo.util.subsystems import ParamSweeper
from pyomo.contrib.incidence_analysis import (
    solve_strongly_connected_components,
)

from idaes.core import FlowsheetBlock

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_large_residuals,
    fixed_variables_set,
    activated_constraints_set,
)

from idaes.core.solvers import get_solver

from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.gas_phase_thermo import (
    GasPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.hetero_reactions import (
    HeteroReactionParameterBlock,
)


# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.fixture(scope="class")
def rxn_prop():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # Set up thermo props and reaction props
    m.fs.solid_properties = SolidPhaseParameterBlock()
    m.fs.solid_state_block = m.fs.solid_properties.build_state_block(
        [0], parameters=m.fs.solid_properties, defined_state=True
    )

    m.fs.gas_properties = GasPhaseParameterBlock()
    m.fs.gas_state_block = m.fs.gas_properties.build_state_block(
        [0], parameters=m.fs.gas_properties, defined_state=True
    )

    m.fs.reactions = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_properties,
        gas_property_package=m.fs.gas_properties,
    )
    m.fs.unit = m.fs.reactions.reaction_block_class(
        [0],
        parameters=m.fs.reactions,
        solid_state_block=m.fs.solid_state_block,
        gas_state_block=m.fs.gas_state_block,
        has_equilibrium=False,
    )

    # Fix required variables to make reaction model square
    # (gas mixture and component densities,
    # solid particle porosity, density and component fractions)
    m.fs.gas_state_block[0].dens_mol.fix(10)
    m.fs.gas_state_block[0].dens_mol_comp.fix(10)
    m.fs.solid_state_block[0].particle_porosity.fix(0.23)
    m.fs.solid_state_block[0].mass_frac_comp["Fe2O3"].fix(0.244)
    m.fs.solid_state_block[0].mass_frac_comp["Fe3O4"].fix(0.202)
    m.fs.solid_state_block[0].mass_frac_comp["Al2O3"].fix(0.554)
    m.fs.solid_state_block[0].dens_mass_skeletal.fix(3125)
    m.fs.solid_state_block[0].temperature.fix(1183.15)  # K

    return m


@pytest.mark.unit
def test_build_reaction_block(rxn_prop):
    assert isinstance(rxn_prop.fs.unit[0].k_rxn, Var)
    assert isinstance(rxn_prop.fs.unit[0].OC_conv, Var)
    assert isinstance(rxn_prop.fs.unit[0].OC_conv_temp, Var)
    assert isinstance(rxn_prop.fs.unit[0].reaction_rate, Var)


@pytest.mark.unit
def test_setInputs_reaction_block(rxn_prop):
    assert degrees_of_freedom(rxn_prop.fs.unit[0]) == 0


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize(rxn_prop):
    orig_fixed_vars = fixed_variables_set(rxn_prop)
    orig_act_consts = activated_constraints_set(rxn_prop)

    rxn_prop.fs.unit.initialize()

    assert degrees_of_freedom(rxn_prop) == 0

    fin_fixed_vars = fixed_variables_set(rxn_prop)
    fin_act_consts = activated_constraints_set(rxn_prop)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve(rxn_prop):

    assert hasattr(rxn_prop.fs.unit[0], "k_rxn")
    assert hasattr(rxn_prop.fs.unit[0], "OC_conv")
    assert hasattr(rxn_prop.fs.unit[0], "reaction_rate")

    results = solver.solve(rxn_prop.fs.unit[0])

    # Check for optimal solution
    assert check_optimal_termination(results)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solution(rxn_prop):
    assert pytest.approx(1, rel=1e-5) == rxn_prop.fs.unit[0].k_rxn["R1"].value
    assert pytest.approx(0, rel=1e-5) == rxn_prop.fs.unit[0].OC_conv.value
    assert pytest.approx(0, rel=1e-5) == rxn_prop.fs.unit[0].reaction_rate["R1"].value


@pytest.mark.component
def test_units_consistent(rxn_prop):
    assert_units_consistent(rxn_prop)


@pytest.mark.unit
def test_state_vars():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.solid_properties = SolidPhaseParameterBlock()
    m.fs.solid_state = m.fs.solid_properties.build_state_block(
        [0], parameters=m.fs.solid_properties, defined_state=True
    )

    m.fs.gas_properties = GasPhaseParameterBlock()
    m.fs.gas_state = m.fs.gas_properties.build_state_block(
        [0], parameters=m.fs.gas_properties, defined_state=True
    )

    m.fs.reaction_properties = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_properties,
        gas_property_package=m.fs.gas_properties,
    )
    m.fs.reaction_block = m.fs.reaction_properties.reaction_block_class(
        [0],
        parameters=m.fs.reaction_properties,
        solid_state_block=m.fs.solid_state,
        gas_state_block=m.fs.gas_state,
        has_equilibrium=False,
    )

    for var in m.fs.gas_state[0].define_state_vars().values():
        var.fix()
    for var in m.fs.solid_state[0].define_state_vars().values():
        var.fix()

    # Note that these checks are necessary to trigger the construction of
    # the reaction block variables
    assert isinstance(m.fs.reaction_block[0].k_rxn, Var)
    assert isinstance(m.fs.reaction_block[0].OC_conv, Var)
    assert isinstance(m.fs.reaction_block[0].OC_conv_temp, Var)
    assert isinstance(m.fs.reaction_block[0].reaction_rate, Var)

    assert degrees_of_freedom(m) == 0

    rxn_vars = list(m.fs.reaction_block[0].component_data_objects(Var))
    rxn_cons = list(m.fs.reaction_block[0].component_data_objects(Constraint))
    assert len(rxn_vars) == 4
    assert len(rxn_cons) == 4


@pytest.mark.unit
class TestProperties(unittest.TestCase):
    def _make_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.solid_properties = SolidPhaseParameterBlock()
        m.fs.solid_state = m.fs.solid_properties.build_state_block(
            parameters=m.fs.solid_properties, defined_state=True
        )

        m.fs.gas_properties = GasPhaseParameterBlock()
        m.fs.gas_state = m.fs.gas_properties.build_state_block(
            parameters=m.fs.gas_properties, defined_state=True
        )

        m.fs.reaction_properties = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_properties,
            gas_property_package=m.fs.gas_properties,
        )
        m.fs.reaction_block = m.fs.reaction_properties.reaction_block_class(
            parameters=m.fs.reaction_properties,
            solid_state_block=m.fs.solid_state,
            gas_state_block=m.fs.gas_state,
            has_equilibrium=False,
        )

        for var in m.fs.gas_state.define_state_vars().values():
            var.fix()
        for var in m.fs.solid_state.define_state_vars().values():
            var.fix()
        return m

    def test_k_rxn(self):
        m = self._make_model()
        rxn_block = m.fs.reaction_block

        n_scen = 4
        mols = pyunits.mol / pyunits.s
        kgs = pyunits.kg / pyunits.s
        K = pyunits.K
        bar = pyunits.bar
        state_values = {
            "gas_state.flow_mol": [1.0 * mols] * n_scen,
            "solid_state.flow_mass": [1.0 * kgs] * n_scen,
            "gas_state.temperature": [1000.0 * K, 1100.0 * K, 1200.0 * K, 1300.0 * K],
            "solid_state.temperature": [1000.0 * K, 1100.0 * K, 1200.0 * K, 1300.0 * K],
            "gas_state.pressure": [1.0 * bar] * n_scen,
            "solid_state.particle_porosity": [0.27] * n_scen,
            "gas_state.mole_frac_comp[O2]": [0.25] * n_scen,
            "gas_state.mole_frac_comp[N2]": [0.25] * n_scen,
            "gas_state.mole_frac_comp[H2O]": [0.25] * n_scen,
            "gas_state.mole_frac_comp[CO2]": [0.25] * n_scen,
            "solid_state.mass_frac_comp[Fe2O3]": [1.0 / 3.0] * n_scen,
            "solid_state.mass_frac_comp[Fe3O4]": [1.0 / 3.0] * n_scen,
            "solid_state.mass_frac_comp[Al2O3]": [1.0 / 3.0] * n_scen,
        }
        state_values = ComponentMap(
            (m.fs.find_component(name), values) for name, values in state_values.items()
        )

        # Units of k_rxn are "non-physical" si units, chosen to be
        # consistent with the reaction rate rule.
        k_rxn_units = pyunits.m / pyunits.s
        target_values = {
            "reaction_block.k_rxn[R1]": [
                5.7556e-5 * k_rxn_units,
                6.7076e-5 * k_rxn_units,
                7.6203e-5 * k_rxn_units,
                8.4888e-5 * k_rxn_units,
            ],
        }
        target_values = ComponentMap(
            (m.fs.find_component(name), values)
            for name, values in target_values.items()
        )

        assert degrees_of_freedom(m.fs) == 0

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, outputs in param_sweeper:
                solve_strongly_connected_components(m.fs)

                # Make sure property equalites have been converged
                assert number_large_residuals(m.fs, tol=1e-8) == 0

                # Sanity checks that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in outputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_oc_conv(self):
        m = self._make_model()
        rxn_block = m.fs.reaction_block

        n_scen = 3
        mols = pyunits.mol / pyunits.s
        kgs = pyunits.kg / pyunits.s
        K = pyunits.K
        bar = pyunits.bar
        state_values = {
            "gas_state.flow_mol": [1.0 * mols] * n_scen,
            "solid_state.flow_mass": [1.0 * kgs] * n_scen,
            "gas_state.temperature": [1273.0 * K] * n_scen,
            "solid_state.temperature": [1273.0 * K] * n_scen,
            "gas_state.pressure": [1.0 * bar] * n_scen,
            "solid_state.particle_porosity": [0.27] * n_scen,
            "gas_state.mole_frac_comp[O2]": [0.25] * n_scen,
            "gas_state.mole_frac_comp[N2]": [0.25] * n_scen,
            "gas_state.mole_frac_comp[H2O]": [0.25] * n_scen,
            "gas_state.mole_frac_comp[CO2]": [0.25] * n_scen,
            "solid_state.mass_frac_comp[Fe2O3]": [2 / 3, 0.0, 1 / 3],
            "solid_state.mass_frac_comp[Fe3O4]": [0.0, 2 / 3, 1 / 3],
            "solid_state.mass_frac_comp[Al2O3]": [1 / 3] * n_scen,
        }
        state_values = ComponentMap(
            (m.fs.find_component(name), values) for name, values in state_values.items()
        )

        target_values = {
            "reaction_block.OC_conv": [
                1.0,
                0.0,
                0.4915,
            ],
            "reaction_block.OC_conv_temp": [
                0.000,
                1.0,
                0.6371,
            ],
        }
        target_values = ComponentMap(
            (m.fs.find_component(name), values)
            for name, values in target_values.items()
        )

        assert degrees_of_freedom(m.fs) == 0

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, outputs in param_sweeper:
                solve_strongly_connected_components(m.fs)

                # Make sure property equalites have been converged
                assert number_large_residuals(m.fs, tol=1e-8) == 0

                # Sanity checks that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in outputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_reaction_rate(self):
        m = self._make_model()
        rxn_block = m.fs.reaction_block

        n_scen = 9
        mols = pyunits.mol / pyunits.s
        kgs = pyunits.kg / pyunits.s
        K = pyunits.K
        bar = pyunits.bar
        state_values = {
            "gas_state.flow_mol": [1.0 * mols] * n_scen,
            "solid_state.flow_mass": [1.0 * kgs] * n_scen,
            "gas_state.temperature": [1273.0 * K] * n_scen,
            "solid_state.temperature": [
                1273.0 * K,
                1273.0 * K,
                1273.0 * K,
                1273.0 * K,
                1273.0 * K,
                1273.0 * K,
                1100.0 * K,
                1200.0 * K,
                1300.0 * K,
            ],
            "gas_state.pressure": [1.0 * bar] * n_scen,
            "solid_state.particle_porosity": [0.27] * n_scen,
            "gas_state.mole_frac_comp[O2]": [
                1.0,
                0.7,
                0.0,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
            "gas_state.mole_frac_comp[N2]": [
                0.0,
                0.1,
                1 / 3,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
            "gas_state.mole_frac_comp[H2O]": [
                0.0,
                0.1,
                1 / 3,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
            "gas_state.mole_frac_comp[CO2]": [
                0.0,
                0.1,
                1 / 3,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
            "solid_state.mass_frac_comp[Fe2O3]": [
                1 / 3,
                1 / 3,
                1 / 3,
                2 / 3,
                0.0,
                1 / 3,
                1 / 3,
                1 / 3,
                1 / 3,
            ],
            "solid_state.mass_frac_comp[Fe3O4]": [
                1 / 3,
                1 / 3,
                1 / 3,
                0.0,
                2 / 3,
                1 / 3,
                1 / 3,
                1 / 3,
                1 / 3,
            ],
            "solid_state.mass_frac_comp[Al2O3]": [1 / 3] * n_scen,
        }
        state_values = ComponentMap(
            (m.fs.find_component(name), values) for name, values in state_values.items()
        )

        molm3s = pyunits.mol / pyunits.m**3 / pyunits.s
        target_values = {
            "reaction_block.reaction_rate[R1]": [
                351.367 * molm3s,
                245.957 * molm3s,
                3.7189815316196473e-07 * molm3s,
                0.000 * molm3s,
                271.731 * molm3s,
                87.842 * molm3s,
                71.344 * molm3s,
                81.051 * molm3s,
                90.288 * molm3s,
            ],
        }
        target_values = ComponentMap(
            (m.fs.find_component(name), values)
            for name, values in target_values.items()
        )

        assert degrees_of_freedom(m.fs) == 0

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, outputs in param_sweeper:
                solve_strongly_connected_components(m.fs)

                # Make sure property equalites have been converged
                assert number_large_residuals(m.fs, tol=1e-8) == 0

                # Sanity checks that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in outputs.items():
                    if value(val) != 0:
                        # To get around Pyomo issue #1627
                        val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pytest
import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.models.properties import iapws95
from idaes.models_extra.power_generation.unit_models import FWH0D
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
from idaes.models.properties.general_helmholtz import helmholtz_available

# Set up solver
solver = get_solver()


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_fwh_model():
    model = pyo.ConcreteModel()
    model.fs = FlowsheetBlock(
        dynamic=False, default_property_package=iapws95.Iapws95ParameterBlock()
    )
    model.fs.properties = model.fs.config.default_property_package
    model.fs.fwh = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=True,
        property_package=model.fs.properties,
    )

    model.fs.fwh.desuperheat.hot_side_inlet.flow_mol[:].set_value(100)
    model.fs.fwh.desuperheat.hot_side_inlet.pressure.fix(201325)
    model.fs.fwh.desuperheat.hot_side_inlet.enth_mol.fix(60000)
    model.fs.fwh.drain_mix.drain.flow_mol.fix(1)
    model.fs.fwh.drain_mix.drain.pressure.fix(201325)
    model.fs.fwh.drain_mix.drain.enth_mol.fix(20000)
    model.fs.fwh.cooling.cold_side_inlet.flow_mol.fix(400)
    model.fs.fwh.cooling.cold_side_inlet.pressure.fix(101325)
    model.fs.fwh.cooling.cold_side_inlet.enth_mol.fix(3000)
    model.fs.fwh.condense.area.fix(1000)
    model.fs.fwh.condense.overall_heat_transfer_coefficient.fix(100)
    model.fs.fwh.desuperheat.area.fix(1000)
    model.fs.fwh.desuperheat.overall_heat_transfer_coefficient.fix(10)
    model.fs.fwh.cooling.area.fix(1000)
    model.fs.fwh.cooling.overall_heat_transfer_coefficient.fix(10)
    model.fs.fwh.initialize(optarg={"max_iter": 50})

    assert degrees_of_freedom(model) == 0
    assert (
        abs(pyo.value(model.fs.fwh.desuperheat.hot_side_inlet.flow_mol[0]) - 98.335)
        < 0.01
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_fwh_units():
    model = pyo.ConcreteModel()
    model.fs = FlowsheetBlock(
        dynamic=False, default_property_package=iapws95.Iapws95ParameterBlock()
    )
    model.fs.properties = model.fs.config.default_property_package
    model.fs.fwh = FWH0D(
        has_desuperheat=True,
        has_drain_cooling=True,
        has_drain_mixer=True,
        property_package=model.fs.properties,
    )

    assert_units_consistent(model)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for calculating fugacity of Henry's Law components

For now, only mole fraction basis (Kpx) form is fully supported. The remainder
is prototype code
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

from enum import Enum

from pyomo.environ import log, Var

from idaes.models.properties.modular_properties.base.utility import StateIndex
from idaes.core.util.exceptions import ConfigurationError


class HenryType(Enum):
    """
    'Henry Constant' types are numbered 1-50 (i.e. H = conc/pressure)
    'Henry Volatility' types are numbered 51-100 (i.e. K = pressure/conc)
    We use this fact to simplify determining wheterh to multiply or divide
    by Henry's constant
    Any different forms can use values 101+, but will need to add the custom
    code in if branches where necessary
    """

    # TODO: Add more forms as needed
    Hcp = 1
    Kpc = 51
    Hxp = 2
    Kpx = 52
    Dummy = 999  # To test error handling


# Define a method for getting the correct concentration term
# TODO : Look to see if this can be unified with get_concentration_term
# in generic/utilties
def get_henry_concentration_term(blk, henry_dict, log=False):
    if log:
        pre = "log_"
    else:
        pre = ""

    if hasattr(blk.params, "_electrolyte") and blk.params._electrolyte:
        if henry_dict["basis"] == StateIndex.true:
            sub = "_true"
        else:
            sub = "_apparent"
    else:
        sub = ""

    henry_type = henry_dict["type"]
    if henry_type == HenryType.Hcp or henry_type == HenryType.Kpc:
        conc_type = "conc_mol_phase_comp"
    elif henry_type == HenryType.Hxp or henry_type == HenryType.Kpx:
        conc_type = "mole_frac_phase_comp"
    else:
        _raise_henry_type_error(henry_type)

    return getattr(blk, pre + conc_type + sub)


# TODO pressure -> fugacity
# Define a method for returning vapor pressure of Henry components
def henry_pressure(b, p, j, T=None):
    henry_def = b.params.get_component(j).config.henry_component[p]

    if T is None:
        henry = b.henry[p, j]
    else:
        henry = henry_def["method"].return_expression(b, p, j, T)
    # Need to get the appropriate concentration term
    # TODO: Add support for true and apparent bases

    h_conc = get_henry_concentration_term(b, henry_def, log=False)[p, j]

    if henry_def["type"].value <= 50:
        # H = c/P type
        h_press = h_conc / henry
    elif henry_def["type"].value <= 100:
        # K = P/c type
        h_press = h_conc * henry
    else:
        _raise_henry_type_error(henry_def["type"])

    return h_press


# TODO pressure -> fugacity
def log_henry_pressure(b, p, j, T=None):
    henry_def = b.params.get_component(j).config.henry_component[p]

    # TODO: Should use a log henry var/expression
    if T is None:
        henry = b.henry[p, j]
    else:
        henry = henry_def["method"].return_expression(b, p, j, T)

    # Need to get the appropriate concentration term
    # TODO: Add support for true and apparent bases

    h_conc = get_henry_concentration_term(b, henry_def, log=True)[p, j]

    if henry_def["type"].value <= 50:
        # H = c/P type
        log_h_press = h_conc - log(henry)
    elif henry_def["type"].value <= 100:
        # K = P/c type
        log_h_press = h_conc + log(henry)
    else:
        _raise_henry_type_error(henry_def["type"])

    return log_h_press


def henry_equilibrium_ratio(b, p, j):
    """
    Returns vapor/liquid equilibrium mole ratio of Henry component j at the
    temperature and pressure of block b.

    Arguments:
        b: Property block for which this calculation is taking place
        p: Liquid phase for which this ratio is being calculated
        j: Henry component in phase p

    Returns:
        Molar ratio of component j in the vapor phase to that in liquid phase p

    Notes:
        If Henry's law is defined in terms of mole fraction, this method
        gives a meaningful answer whether or not the liquid phase composition
        has been set. If it is defined in terms of concentration, a reasonable
        value is given only if a reasonable liquid phase composition has been
        specified.

        This function returns a value regardless of whether or not block b
        is experiencing phase equilibrium at its current conditions.
    """

    henry_def = b.params.get_component(j).config.henry_component[p]
    henry_constant = b.henry[p, j]
    if henry_def["type"] == HenryType.Hcp:
        henry_constant /= b.dens_mol_phase[p]
    elif henry_def["type"] == HenryType.Kpc:
        henry_constant *= b.dens_mol_phase[p]

    if henry_def["type"].value <= 50:
        # H = c/P type
        return 1 / (henry_constant * b.pressure)
    elif henry_def["type"].value <= 100:
        # K = P/c type
        return henry_constant / b.pressure
    else:
        _raise_henry_type_error(henry_def["type"])


# Define units for Henry's constant
def henry_units(henry_type, units):
    if henry_type == HenryType.Hcp:
        h_units = units.DENSITY_MOLE / units.PRESSURE
    elif henry_type == HenryType.Kpc:
        h_units = units.PRESSURE / units.DENSITY_MOLE
    elif henry_type == HenryType.Hxp:
        h_units = units.PRESSURE**-1
    elif henry_type == HenryType.Kpx:
        h_units = units.PRESSURE
    else:
        _raise_henry_type_error(henry_type)

    return h_units


class ConstantH:
    """Methods for invariant Henry constant"""

    @staticmethod
    def build_parameters(cobj, p, h_type):
        b = cobj.parent_block()
        units = b.get_metadata().derived_units
        h_units = henry_units(h_type, units)

        cobj.add_component(
            "henry_ref_" + p,
            Var(
                initialize=cobj.config.parameter_data["henry_ref"][p],
                doc="Henry coefficient (mole fraction basis) at reference "
                "state for phase " + p,
                units=h_units,
            ),
        )

    @staticmethod
    def return_expression(b, p, j, T=None):
        cobj = b.params.get_component(j)
        H = getattr(cobj, "henry_ref_" + p)

        return H

    # TODO: Need a return log expression method too

    @staticmethod
    def dT_expression(b, p, j, T=None):
        return 0


def _raise_henry_type_error(henry_type):
    raise ConfigurationError(f"Unrecognized value for HenryType {henry_type}")
